@@ -1,204 +1,20 @@
 """
-A class that generates a training, validation, and test set from a set of hyperparameters.
+Group together and name your training, validation, and test sets.
 
-The :py:class:`DataBlob` and :py:class:`DataFrameDataBlob`
-classes exist to help you group together and name your
-training, validation, and test sets and the hyperparameters
-to generate them.
+The classes in this module are used to group together
+data into training, validation, and test sets used for
+training machine learning models. We also record the
+hyperparameters used to process the dataset.
 
-An introduction to :py:class:`DataBlob`
----------------------------------------
-
-Create a subclass of :py:class:`DataBlob` named after
-the category of dataset you are creating. Inside your
-subclass, define a Python dataclass named ``Hyperparams``
-and override the methods :py:meth:`DataBlob.set_training`,
-:py:meth:`DataBlob.set_validation`, and :py:meth:`DataBlob.set_test`.
-
-Each one of these overrides should create a *new*
-:py:class:`tf.data.Dataset` instance representing your
-training, validation, and test sets. All of this
-looks like something below.
-
->>> import tensorflow as tf
->>> import scalarstop as sp
->>>
->>> class MyDataBlob(sp.DataBlob):
-...
-...     @sp.dataclass
-...     class Hyperparams(sp.HyperparamsType):
-...             cols: int
-...
-...     def _data(self):
-...             x = tf.random.uniform(shape=(10, self.hyperparams.cols))
-...             y = tf.round(tf.random.uniform(shape=(10,1)))
-...             return tf.data.Dataset.zip((
-...                     tf.data.Dataset.from_tensor_slices(x),
-...                     tf.data.Dataset.from_tensor_slices(y),
-...             ))
-...
-...     def set_training(self):
-...         return self._data()
-...
-...     def set_validation(self):
-...         return self._data()
-...
-...     def set_test(self):
-...         return self._data()
->>>
-
-In our above example, our training, validation, and test sets
-are created with the exact same code. In practice, you'll
-be creating them with different inputs.
-
-Now we create an instance of our subclass so we can start
-using it.
-
->>> datablob = MyDataBlob(hyperparams=dict(cols=3))
->>> datablob
-<sp.DataBlob MyDataBlob-bn5hpc7ueo2uz7as1747tetn>
-
-:py:class:`DataBlob` instances are given a unique name
-by hashing together the class name with the instance's
-hyperparameters.
-
->>> datablob.name
-'MyDataBlob-bn5hpc7ueo2uz7as1747tetn'
->>> datablob.group_name
-'MyDataBlob'
->>> datablob.hyperparams
-MyDataBlob.Hyperparams(cols=3)
->>> sp.enforce_dict(datablob.hyperparams)
-{'cols': 3}
-
-The methods :py:meth:`DataBlob.set_training`,
-:py:meth:`DataBlob.set_validation`, and :py:meth:`DataBlob.set_test`
-are responsible for creating *new* instances of
-:py:class:`tf.data.Dataset` objects.
-
-We save exactly one instance of each :py:class:`tf.data.Dataset`
-pipeline in the properties :py:attr:`DataBlob.training`,
-:py:attr:`DataBlob.validation`, and :py:attr:`DataBlob.test`.
-
->>> datablob.training
-<ZipDataset shapes: ((3,), (1,)), types: (tf.float32, tf.float32)>
->>> datablob.validation
-<ZipDataset shapes: ((3,), (1,)), types: (tf.float32, tf.float32)>
->>> datablob.test
-<ZipDataset shapes: ((3,), (1,)), types: (tf.float32, tf.float32)>
-
-:py:class:`DataBlob` objects have some methods for applying
-:py:mod:`tf.data` transformations to the training, validation, and
-test sets at the same time.
-
-:py:meth:`DataBlob.batch` will batch the training, validation,
-and test sets at the same time. If you call
-:py:meth:`DataBlob.batch` with the keyword argument
-``with_tf_distribute=True``, your input batch size will be
-multiplied by the number of replicas in your :py:mod:`tf.distribute`
-strategy.
-
-:py:meth:`DataBlob.cache` will cache the training, validation,
-and test sets in memory once you iterate over them. This is
-useful if your :py:class:`tf.data.Dataset` are doing something
-computationally expensive each time you iterate over them.
-
-:py:meth:`DataBlob.save` saves the training, validation, and test
-sets to a path on the filesystem. This can be loaded back with
-the classmethod :py:meth:`DataBlob.load_from_directory`.
-
->>> import os
->>> import tempfile
->>> tempdir = tempfile.TemporaryDirectory()
->>> datablob = datablob.save(tempdir.name)
->>> os.listdir(tempdir.name)
-['MyDataBlob-bn5hpc7ueo2uz7as1747tetn']
-
->>> my_datablob_path = os.path.join(tempdir.name, datablob.name)
->>> loaded_datablob = MyDataBlob.load_from_directory(my_datablob_path)
->>> loaded_datablob
-<sp.DataBlob MyDataBlob-bn5hpc7ueo2uz7as1747tetn>
-
-
-An introduction to :py:class:`DataFrameDataBlob`
-------------------------------------------------
-:py:class:`DataBlob` is useful when you want to manually define your
-:py:mod:`tf.data` pipelines and their input tensors. But it is likely
-that your training, validation, and test set inputs are already
-in a data structure like a :py:class:`pandas.DataFrame`.
-
-In that case, you can subclass a :py:class:`DataFrameDataBlob`.
-In your subclass, you can use class attributes to specify
-
-* the :py:class:`pandas.DataFrame` columns to use as the data samples and data labels
-* the percentage of the :py:class:`pandas.DataFrame` to split into the training,
-  validation, and test sets.
-
-Here is how to set the class attributes. These are the default values, by the way.
-
->>> class MyDataFrameDataBlob(sp.DataFrameDataBlob):
-...    samples_column: str = "samples"
-...    labels_column: str = "labels"
-...    training_fraction: float = 0.6
-...    validation_fraction: float = 0.2
-
-The above configuration will split the first 60% of your
-:py:class:`pandas.DataFrame` as the input tensors for your training set,
-the next 20% as the input tensors for your validation set, and the
-remainder of the :py:class:`~pandas.DataFrame` for your test set.
-
-But we are not done yet. In your :py:class:`DataFrameDataBlob` subclass,
-override the :py:meth:`DataFrameDataBlob.set_dataframe` method to
-create a new instance of the :py:class:`pandas.DataFrame` that
-describes all of the samples and labels in your training, validation,
-and test sets.
-
-Then, override :py:meth:`DataFrameDataBlob.transform` with a function
-that transforms an arbitrary :py:class:`pandas.DataFrame` into
-a :py:class:`tf.data.Dataset`. The :py:meth:`DataFrameDataBlob.transform`
-method is called to generate the :py:mod:`tf.data` pipelines
-at :py:attr:`DataFrameDataBlob.training`, :py:attr:`DataFrameDataBlob.validation`,
-and :py:attr:`DataFrameDataBlob.test`.
-
->>> import pandas as pd
->>> import tensorflow as tf
->>> import scalarstop as sp
->>>
->>> class MyDataFrameDataBlob(sp.DataFrameDataBlob):
-...    samples_column: str = "samples"
-...    labels_column: str = "labels"
-...    training_fraction: float = 0.6
-...    validation_fraction: float = 0.2
-...
-...    @sp.dataclass
-...    class Hyperparams(sp.HyperparamsType):
-...        length: int = 0
-...
-...    def set_dataframe(self):
-...        samples = list(range(self.hyperparams.length))
-...        labels = list(range(self.hyperparams.length))
-...        return pd.DataFrame(dict(samples=samples, labels=labels))
-...
-...    def transform(self, dataframe: pd.DataFrame):
-...        return tf.data.Dataset.zip((
-...                tf.data.Dataset.from_tensor_slices(dataframe[self.samples_column]),
-...                tf.data.Dataset.from_tensor_slices(dataframe[self.labels_column]),
-...        ))
-
->>> datablob2 = MyDataFrameDataBlob(hyperparams=dict(length=10))
-
-And you can use the resulting object in all of the same ways as
-we've demonstrated with :py:class:`DataBlob` subclass instances above.
-
-(and now let's clean up the temporary directory from above)
-
->>> tempdir.cleanup()
+The :py:class:`DataBlob` subclass name and hyperparameters
+are used to create a unique content-addressable name
+that makes it easy to keep track of many datasets at once.
 """
 
 import errno
 import json
 import os
-from typing import Any, Mapping, Optional, Union
+from typing import Any, Mapping, Optional, Type, Union
 
 import pandas as pd
 import tensorflow as tf
@@ -215,7 +31,14 @@ from scalarstop.exceptions import (
     IsNotImplemented,
     TensorFlowDatasetNotFound,
 )
-from scalarstop.hyperparams import HyperparamsType, hash_hyperparams, init_hyperparams
+from scalarstop.hyperparams import (
+    AppendHyperparamsType,
+    HyperparamsType,
+    NestedHyperparamsType,
+    enforce_dict,
+    hash_hyperparams,
+    init_hyperparams,
+)
 
 _METADATA_JSON_FILENAME = "metadata.json"
 _METADATA_PICKLE_FILENAME = "metadata.pickle"
@@ -228,7 +51,7 @@ _TFDATA_DIRECTORY_NAME = "tfdata"
 _SUBTYPES = (_SUBTYPE_TRAINING, _SUBTYPE_VALIDATION, _SUBTYPE_TEST)
 
 
-def _load_tfdata_dataset(path, element_spec=None):
+def _load_tfdata_dataset(path, element_spec=None) -> tf.data.Dataset:
     """
     Load a :py:class:`tf.data.Dataset` from a filesystem path.
 
@@ -262,9 +85,147 @@ def _load_tfdata_dataset(path, element_spec=None):
 
 
 class DataBlob:
-    """Subclass this and override :py:meth:`~DataBlob.set_training`, :py:meth:`~DataBlob.set_validation`, and :py:meth:`~DataBlob.set_test` to create :py:class:`tf.data.Dataset` pipelines."""  # pylint: disable=line-too-long
+    """
+    Subclass this to group your training, validation, and test sets for training machine learning models.
 
-    Hyperparams = HyperparamsType
+    Here is how to use :py:class:`DataBlob` to group your training,
+    validation, and test sets:
+
+    1. Subclass :py:class:`DataBlob` with a class name that describes
+       your dataset in general. In this example, we'll use
+       ``MyDataBlob`` as the class name.
+
+    2. Define a dataclass using the ``@sp.dataclass`` decorator at
+       ``MyDataBlob.Hyperparams``. We'll define an *instance* of
+       this dataclass at ``MyDataBlob.hyperparams``. This describes
+       the hyperparameters involved in processing your dataset.
+
+    3. Override the methods :py:meth:`DataBlob.set_training`,
+       :py:meth:`DataBlob.set_validation`, and :py:meth:`DataBlob.set_test`
+       to generate :py:class:`tf.data.Dataset` pipelines
+       representing your training, validation, and test sets.
+
+    Those three steps roughly look like:
+
+    >>> import tensorflow as tf
+    >>> import scalarstop as sp
+    >>>
+    >>> class MyDataBlob(sp.DataBlob):
+    ...
+    ...     @sp.dataclass
+    ...     class Hyperparams(sp.HyperparamsType):
+    ...             cols: int
+    ...
+    ...     def _data(self):
+    ...             x = tf.random.uniform(shape=(10, self.hyperparams.cols))
+    ...             y = tf.round(tf.random.uniform(shape=(10,1)))
+    ...             return tf.data.Dataset.zip((
+    ...                     tf.data.Dataset.from_tensor_slices(x),
+    ...                     tf.data.Dataset.from_tensor_slices(y),
+    ...             ))
+    ...
+    ...     def set_training(self):
+    ...         return self._data()
+    ...
+    ...     def set_validation(self):
+    ...         return self._data()
+    ...
+    ...     def set_test(self):
+    ...         return self._data()
+    >>>
+
+    In our above example, our training, validation, and test sets
+    are created with the exact same code. In practice, you'll
+    be creating them with different inputs.
+
+    Now we create an instance of our subclass so we can start
+    using it.
+
+    >>> datablob = MyDataBlob(hyperparams=dict(cols=3))
+    >>> datablob
+    <sp.DataBlob MyDataBlob-bn5hpc7ueo2uz7as1747tetn>
+
+    :py:class:`DataBlob` instances are given a unique name
+    by hashing together the class name with the instance's
+    hyperparameters.
+
+    >>> datablob.name
+    'MyDataBlob-bn5hpc7ueo2uz7as1747tetn'
+    >>>
+    >>> datablob.group_name
+    'MyDataBlob'
+    >>>
+    >>> datablob.hyperparams
+    MyDataBlob.Hyperparams(cols=3)
+    >>>
+    >>> sp.enforce_dict(datablob.hyperparams)
+    {'cols': 3}
+
+    We save exactly one instance of each :py:class:`tf.data.Dataset`
+    pipeline in the properties :py:attr:`DataBlob.training`,
+    :py:attr:`DataBlob.validation`, and :py:attr:`DataBlob.test`.
+
+    >>> datablob.training
+    <ZipDataset shapes: ((3,), (1,)), types: (tf.float32, tf.float32)>
+    >>>
+    >>> datablob.validation
+    <ZipDataset shapes: ((3,), (1,)), types: (tf.float32, tf.float32)>
+    >>>
+    >>> datablob.test
+    <ZipDataset shapes: ((3,), (1,)), types: (tf.float32, tf.float32)>
+
+    :py:class:`DataBlob` objects have some methods for applying
+    :py:mod:`tf.data` transformations to the training, validation, and
+    test sets at the same time:
+
+    * **Batching.** :py:meth:`DataBlob.batch` will batch the training, validation,
+      and test sets at the same time. If you call
+      :py:meth:`DataBlob.batch` with the keyword argument
+      ``with_tf_distribute=True``, your input batch size will be
+      multiplied by the number of replicas in your :py:mod:`tf.distribute`
+      strategy.
+
+    * **Caching.** :py:meth:`DataBlob.cache` will cache the training, validation,
+      and test sets in memory once you iterate over them. This is
+      useful if your :py:class:`tf.data.Dataset` are doing something
+      computationally expensive each time you iterate over them.
+
+    * **Saving/loading to/from the filesystem.** :py:meth:`DataBlob.save`
+      saves the training, validation, and test
+      sets to a path on the filesystem. This can be loaded back with
+      the classmethod :py:meth:`DataBlob.from_exact_path`.
+
+    >>> import os
+    >>> import tempfile
+    >>> tempdir = tempfile.TemporaryDirectory()
+    >>>
+    >>> datablob = datablob.save(tempdir.name)
+    >>>
+    >>> os.listdir(tempdir.name)
+    ['MyDataBlob-bn5hpc7ueo2uz7as1747tetn']
+
+    >>> path = os.path.join(tempdir.name, datablob.name)
+    >>> loaded_datablob = MyDataBlob.from_exact_path(path)
+    >>> loaded_datablob
+    <sp.DataBlob MyDataBlob-bn5hpc7ueo2uz7as1747tetn>
+
+    Alternatively, if you have the hyperparameters of the
+    :py:class:`DataBlob` but not the name, you can use the
+    classmethod :py:meth:`DataBlob.from_filesystem`.
+
+    >>> loaded_datablob_2 = MyDataBlob.from_filesystem(
+    ...    hyperparams=dict(cols=3),
+    ...    datablobs_directory=tempdir.name,
+    ... )
+    >>> loaded_datablob_2
+    <sp.DataBlob MyDataBlob-bn5hpc7ueo2uz7as1747tetn>
+
+    (and now let's clean up the temporary directory from above)
+
+    >>> tempdir.cleanup()
+    """  # pylint: disable=line-too-long
+
+    Hyperparams: Type[HyperparamsType] = HyperparamsType
     hyperparams: HyperparamsType
 
     _training: Optional[tf.data.Dataset] = None
@@ -275,45 +236,99 @@ class DataBlob:
     _group_name: Optional[str] = None
 
     def __init__(
-        self, *, hyperparams: Optional[Union[Mapping[str, Any], HyperparamsType]] = None
+        self,  # pylint: disable=unused-argument
+        *,
+        hyperparams: Optional[Union[Mapping[str, Any], HyperparamsType]] = None,
+        **kwargs,
     ):
         self.hyperparams = init_hyperparams(
-            self=self,
+            class_name=self.__class__.__name__,
             hyperparams=hyperparams,
             hyperparams_class=self.Hyperparams,
         )
 
-    @staticmethod
-    def load_from_directory(this_dataset_directory) -> "DataBlob":
-        """Load a :py:class:`DataBlob` from a directory on the filesystem."""
-        return _LoadDataBlob.load_from_directory(
-            this_dataset_directory=this_dataset_directory
+    @classmethod
+    def from_filesystem(
+        cls,
+        *,
+        hyperparams: Optional[Union[Mapping[str, Any], HyperparamsType]] = None,
+        datablobs_directory: str,
+    ):
+        """
+        Load a :py:class:`DataBlob` from the filesystem, calculating the
+        filename from the hyperparameters.
+
+        Args:
+            hyperparams: The hyperparameters of the model that we want to load.
+
+            datablobs_directory: The parent directory of all of your saved
+                :py:class:`DataBlob` s. The exact filename is calculated
+                from the class name and hyperparams.
+        """
+        the_hyperparams = init_hyperparams(
+            class_name=cls.__name__,
+            hyperparams=hyperparams,
+            hyperparams_class=cls.Hyperparams,
         )
+        group_name = cls._make_group_name()
+        name = cls._make_name(group_name=group_name, hyperparams=the_hyperparams)
+        path = os.path.join(datablobs_directory, name)
+        return cls.from_exact_path(path)
+
+    @classmethod
+    def from_filesystem_or_new(
+        cls,
+        *,
+        hyperparams: Optional[Union[Mapping[str, Any], HyperparamsType]] = None,
+        datablobs_directory: str,
+        **kwargs,
+    ):
+        """Load a :py:class:`DataBlob` from the filesystem, calculating the
+        filename from the hyperparameters. Create a new :py:class:`DataBlob`
+        if we cannot find a saved one on the filesystem.
+
+        Args:
+            hyperparams: The hyperparameters of the model that we want to load.
+
+            datablobs_directory: The parent directory of all of your saved
+                :py:class:`DataBlob` s. The exact filename is calculated
+                from the class name and hyperparams.
+
+            **kwargs: Other keyword arguments that you need to pass to
+                your ``__init__()``.
+        """
+        try:
+            return cls.from_filesystem(
+                hyperparams=hyperparams, datablobs_directory=datablobs_directory
+            )
+        except DataBlobNotFound:
+            return cls(hyperparams=hyperparams, **kwargs)
+
+    @staticmethod
+    def from_exact_path(path: str) -> "DataBlob":
+        """Load a :py:class:`DataBlob` from a directory on the filesystem."""
+        return _LoadDataBlob.from_exact_path(path=path)
 
     def __repr__(self) -> str:
         return f"<sp.DataBlob {self.name}>"
 
+    @staticmethod
+    def _make_name(*, group_name: str, hyperparams) -> str:
+        """Compute what this :py:class:`DataBlob` 's name should be."""
+        return "-".join((group_name, hash_hyperparams(hyperparams)))
+
+    @classmethod
+    def _make_group_name(cls) -> str:
+        """Compute what this :py:class:`DataBlob` 's group name should be."""
+        return cls.__name__
+
     @property
     def name(self) -> str:
-        """
-        The name of this specific dataset.
-
-        If you intend on overriding this method, make sure that
-        instances of the same class with different hyperparameters
-        will have different names.
-
-        However, if you have additional parameters to your
-        subclass's ``__init__()`` or configuration that are not
-        hyperparameters to your :py:class:`DataBlob`, then they
-        should not change the name. For example, you might want to
-        pass a database connection string or a filesystem path
-        to your :py:class:`DataBlob` subclass. Other users of your
-        subclass might have different values for those parameters,
-        but they don't change the actual dataset itself--and
-        therefore should not change the :py:class:`DataBlob` name.
-        """
+        """The name of this specific dataset."""
         if self._name is None:
-            self._name = "-".join((self.group_name, hash_hyperparams(self.hyperparams)))
+            self._name = self._make_name(
+                group_name=self.group_name, hyperparams=self.hyperparams
+            )
         return self._name
 
     @property
@@ -327,7 +342,7 @@ class DataBlob:
         share the same code but have different hyperparameters.
         """
         if self._group_name is None:
-            self._group_name = self.__class__.__name__
+            self._group_name = self._make_group_name()
         return self._group_name
 
     def set_training(self) -> tf.data.Dataset:
@@ -373,28 +388,38 @@ class DataBlob:
         """Cache this :py:class:`DataBlob` into memory before iterating over it."""
         return _CacheDataBlob(wraps=self)
 
-    def save_hook(self, *, subtype, path) -> None:  # pylint: disable=unused-argument
+    def save_hook(  # pylint: disable=unused-argument
+        self, *, subtype: str, path: str
+    ) -> None:
         """
         Override this method to run additional code when saving this
         :py:class:`DataBlob` to disk.
         """
         return None
 
-    def save(self, dataset_directory) -> "DataBlob":
+    def save(
+        self, datablobs_directory: str, *, ignore_existing: bool = False
+    ) -> "DataBlob":
         """
         Save this :py:class:`DataBlob` to disk.
 
         Args:
-            dataset_directory: The directory where you plan on storing all of your
+            datablobs_directory: The directory where you plan on storing all of your
                 DataBlobs. This method will save this :py:class:`DataBlob` in a subdirectory
-                of ``dataset_directory`` with same name as :py:attr:`DataBlob.name`.
+                of ``datablobs_directory`` with same name as :py:attr:`DataBlob.name`.
+
+            ignore_existing: Set this to ``True`` to ignore if
+                there is already a :py:class:`DataBlob` at the given
+                path.
 
         Returns:
             Return ``self``, enabling you to place this call in a chain.
         """
         # Begin writing our hyperparameters, dataframes, tfdata, and element spec.
-        final_datablob_path = os.path.join(dataset_directory, self.name)
+        final_datablob_path = os.path.join(datablobs_directory, self.name)
         if os.path.exists(final_datablob_path):
+            if ignore_existing:
+                return self
             raise FileExists(
                 f"File or directory already exists at path {final_datablob_path}"
             )
@@ -483,7 +508,74 @@ class DataBlob:
 
 
 class DataFrameDataBlob(DataBlob):
-    """Subclass this to create a a :py:class:`DataBlob` where the resulting :py:class:`tf.data.Dataset` pipelines are each fed a :py:class:`pandas.DataFrame` of input tensors."""  # pylint: disable=line-too-long
+    """
+    Subclass this to transform a :py:class:`pandas.DataFrame` into your training, validation, and test sets.
+
+    :py:class:`DataBlob` is useful when you want to manually define your
+    :py:mod:`tf.data` pipelines and their input tensors.
+
+    However, if your input tensors are in a fixed-size list or
+    :py:class:`~pandas.DataFrame` that you want to *slice* into
+    a training, validation, and test set, then you might find
+    :py:class:`DataFrameDataBlob` handy.
+
+    Here is how to use it:
+
+    1. Subclass :py:class:`DataFrameDataBlob` with a class name that
+       describes your dataset.
+
+    2. Override :py:meth:`DataFrameDataBlob.set_dataframe` and have
+       it return a *single* :py:class:`~pandas.DataFrame` that contains
+       all of the *inputs* for your training, validation, and
+       test sets. The :py:class:`~pandas.DataFrame` should have
+       one column representing training samples and another
+       column representing training labels.
+
+    3. Override :py:meth:`DataFrameDataBlob.transform` and define
+       a method that transforms an arbitrary :py:class:`~pandas.DataFrame`
+       of *inputs* into a :py:class:`tf.data.Dataset` pipeline
+       that represents the actual dataset needed for training
+       and evaluation.
+
+    We define what fraction of the :py:class:`~pandas.DataFrame` to split
+    with the class attributes :py:attr:`DataFrameDataBlob.training_fraction`
+    and :py:attr:`DataFrameDataBlob.validation_fraction`. By default,
+    60 percent of the :py:class:`~pandas.DataFrame` is marked
+    for the training set, 20 percent for the validation set,
+    and the remainder of the :py:class:`~pandas.DataFrame` for the test set.
+
+    Roughly, this looks like:
+
+    >>> import pandas as pd
+    >>> import tensorflow as tf
+    >>> import scalarstop as sp
+    >>>
+    >>> class MyDataFrameDataBlob(sp.DataFrameDataBlob):
+    ...    samples_column: str = "samples"
+    ...    labels_column: str = "labels"
+    ...    training_fraction: float = 0.6
+    ...    validation_fraction: float = 0.2
+    ...
+    ...    @sp.dataclass
+    ...    class Hyperparams(sp.HyperparamsType):
+    ...        length: int = 0
+    ...
+    ...    def set_dataframe(self):
+    ...        samples = list(range(self.hyperparams.length))
+    ...        labels = list(range(self.hyperparams.length))
+    ...        return pd.DataFrame({self.samples_column: samples, self.labels_column: labels})
+    ...
+    ...    def transform(self, dataframe: pd.DataFrame):
+    ...        return tf.data.Dataset.zip((
+    ...                tf.data.Dataset.from_tensor_slices(dataframe[self.samples_column]),
+    ...                tf.data.Dataset.from_tensor_slices(dataframe[self.labels_column]),
+    ...        ))
+
+    >>> datablob2 = MyDataFrameDataBlob(hyperparams=dict(length=10))
+
+    And you can use the resulting object in all of the same ways as
+    we've demonstrated with :py:class:`DataBlob` subclass instances above.
+    """  # pylint: disable=line-too-long
 
     samples_column: str = "samples"
     labels_column: str = "labels"
@@ -496,11 +588,12 @@ class DataFrameDataBlob(DataBlob):
     _test_dataframe: Optional[pd.DataFrame] = None
 
     @staticmethod
-    def load_from_directory(this_dataset_directory) -> "DataFrameDataBlob":
+    def from_exact_path(
+        path: str,
+    ) -> Union[DataBlob, "DataFrameDataBlob"]:
         """Load a :py:class:`DataFrameDataBlob` from a directory on the filesystem."""
-        return _LoadDataFrameDataBlob.load_from_directory(
-            this_dataset_directory=this_dataset_directory
-        )
+        loaded = _LoadDataFrameDataBlob.from_exact_path(path=path)
+        return loaded
 
     def __repr__(self) -> str:
         return f"<sp.DataFrameDataBlob {self.name}>"
@@ -612,7 +705,7 @@ class DataFrameDataBlob(DataBlob):
     def set_test(self) -> tf.data.Dataset:
         return self.transform(self.test_dataframe)
 
-    def save_hook(self, *, subtype, path) -> None:
+    def save_hook(self, *, subtype: str, path: str) -> None:
         super().save_hook(subtype=subtype, path=path)
         dataframe = getattr(self, subtype + "_dataframe", None)
         dataframe_path = os.path.join(path, _DATAFRAME_FILENAME)
@@ -636,7 +729,7 @@ class _WrapDataBlob(DataBlob):
         self._name = wraps.name
         self._group_name = wraps.group_name
 
-    def __getattribute__(self, key):
+    def __getattribute__(self, key: str):
         """
         Returns attributes on the wrapped object for any key that isn't present on this object.
 
@@ -649,7 +742,7 @@ class _WrapDataBlob(DataBlob):
             wraps = object.__getattribute__(self, "_wraps")
             return wraps.__getattribute__(key)
 
-    def _wrap_tfdata(self, tfdata) -> tf.data.Dataset:
+    def _wrap_tfdata(self, tfdata: tf.data.Dataset) -> tf.data.Dataset:
         raise IsNotImplemented("_WrapDataBlob._wrap_tfdata()")
 
     def set_training(self) -> tf.data.Dataset:
@@ -682,10 +775,185 @@ class _WrapDataBlob(DataBlob):
             self._test = self._wrap_tfdata(self._wraps.test)
         return self._test
 
-    def save_hook(self, *, subtype, path) -> None:
+    def save_hook(self, *, subtype: str, path: str) -> None:
         return self._wraps.save_hook(subtype=subtype, path=path)
 
 
+class AppendDataBlob(DataBlob):
+    """
+    Subclass this to create a new :py:class:`DataBlob` that extends an existing :py:class:`DataBlob`.
+
+    The :py:class:`AppendDataBlob` class is useful when you have an existing
+    :py:class:`DataBlob` or :py:class:`DataFrameDataBlob` with most, but not *all*
+    of the functionality you need. If you are trying to implement multiple
+    data pipelines that share a common compute-intensive first step, you
+    can implement your pipelines as :py:class:`AppendDataBlob` subclasses with
+    the common first step as a :py:class:`DataBlob` that you save and load
+    to/from the filesystem.
+
+    Let's begin by creating a :py:class:`DataBlob` that we will use as a
+    parent for an :py:class:`AppendDataBlob`.
+
+    >>> import tensorflow as tf
+    >>> import scalarstop as sp
+    >>>
+    >>> class MyDataBlob(sp.DataBlob):
+    ...
+    ...     @sp.dataclass
+    ...     class Hyperparams(sp.HyperparamsType):
+    ...             length: int
+    ...
+    ...     def _data(self):
+    ...         length = self.hyperparams.length
+    ...         x = tf.data.Dataset.from_tensor_slices(list(range(0, length)))
+    ...         y = tf.data.Dataset.from_tensor_slices(list(range(length, length * 2)))
+    ...         return tf.data.Dataset.zip((x, y))
+    ...
+    ...     def set_training(self):
+    ...         return self._data()
+    ...
+    ...     def set_validation(self):
+    ...         return self._data()
+    ...
+    ...     def set_test(self):
+    ...         return self._data()
+    >>>
+
+    And then we create an instance of the datablob and save it to
+    the filesystem.
+
+    >>> import os
+    >>> import tempfile
+    >>> tempdir = tempfile.TemporaryDirectory()
+    >>>
+    >>> datablob = MyDataBlob(hyperparams=dict(length=5))
+    >>> datablob
+    <sp.DataBlob MyDataBlob-dac936v7mb1ue9phjp6tc3sb>
+    >>>
+    >>> list(datablob.training.as_numpy_iterator())
+    [(0, 5), (1, 6), (2, 7), (3, 8), (4, 9)]
+    >>>
+    >>> datablob = datablob.save(tempdir.name)
+    >>>
+    >>> os.listdir(tempdir.name)
+    ['MyDataBlob-dac936v7mb1ue9phjp6tc3sb']
+
+    Now, let's say that we want to create an :py:class:`AppendDataBlob`
+    that takes in any input :py:class:`DataBlob` or
+    :py:class:`DataFrameDataBlob` and multiplies every number in every tensor by
+    a constant.
+
+    >>> class MyAppendDataBlob(sp.AppendDataBlob):
+    ...
+    ...     @sp.dataclass
+    ...     class Hyperparams(sp.AppendHyperparamsType):
+    ...          coefficient: int
+    ...
+    ...     hyperparams: "MyAppendDataBlob.Hyperparams"
+    ...
+    ...     def __init__(self, *, parent: sp.DataBlob, hyperparams):
+    ...         hyperparams_dict = sp.enforce_dict(hyperparams)
+    ...         if hyperparams_dict["coefficient"] < 1:
+    ...             raise ValueError("Coefficient is too low.")
+    ...         super().__init__(parent=parent, hyperparams=hyperparams_dict)
+    ...
+    ...     def _wrap_tfdata(self, tfdata: tf.data.Dataset) -> tf.data.Dataset:
+    ...          return tfdata.map(
+    ...              lambda x, y: (
+    ...                  x * self.hyperparams.coefficient,
+    ...                  y * self.hyperparams.coefficient,
+    ...               )
+    ...          )
+    >>>
+    >>> append = MyAppendDataBlob(parent=datablob, hyperparams=dict(coefficient=3))
+    >>> list(append.training.as_numpy_iterator())
+    [(0, 15), (3, 18), (6, 21), (9, 24), (12, 27)]
+
+    (And now let's clean up the temporary directory that we created earlier.)
+
+    >>> tempdir.cleanup()
+    """  # pylint: disable=line-too-long
+
+    Hyperparams: Type[AppendHyperparamsType] = AppendHyperparamsType
+
+    def __init__(
+        self,
+        *,
+        parent: DataBlob,
+        hyperparams: Optional[Union[Mapping[str, Any], HyperparamsType]] = None,
+    ):
+        """
+        Args:
+            parent: The :py:class:`DataBlob` to extend.
+
+            hyperparams: Additional hyperparameters to add on top of the
+                existing hyperparameters from the parent :py:class:`DataBlob`.
+        """
+        super().__init__(
+            hyperparams=dict(
+                parent=NestedHyperparamsType(
+                    name=parent.name,
+                    group_name=parent.group_name,
+                    hyperparams=parent.hyperparams,
+                ),
+                **enforce_dict(hyperparams),
+            )
+        )
+        self._parent = parent
+
+    def __getattribute__(self, key: str):
+        """
+        Returns attributes on the wrapped object for any key that isn't present on this object.
+
+        Args:
+            key: The key to look up on this object (and then on the wrapped object).
+        """
+        try:
+            return object.__getattribute__(self, key)
+        except AttributeError:
+            parent = object.__getattribute__(self, "_parent")
+            return parent.__getattribute__(key)
+
+    @property
+    def parent(self) -> DataBlob:
+        """The parent :py:class:`DataBlob`."""
+        return self._parent
+
+    def _wrap_tfdata(self, tfdata: tf.data.Dataset) -> tf.data.Dataset:
+        """
+        Override this to add something to the training, validation,
+        and test :py:class:`tf.data.Dataset` pipelines.
+        """
+        raise IsNotImplemented("AppendDataBlob._wrap_tfdata()")
+
+    def set_training(self) -> tf.data.Dataset:
+        return self._wrap_tfdata(self.parent.set_training())
+
+    @property
+    def training(self) -> tf.data.Dataset:
+        if self._training is None:
+            self._training = self._wrap_tfdata(self.parent.training)
+        return self._training
+
+    def set_validation(self) -> tf.data.Dataset:
+        return self._wrap_tfdata(self.parent.set_validation())
+
+    @property
+    def validation(self) -> tf.data.Dataset:
+        if self._validation is None:
+            self._validation = self._wrap_tfdata(self.parent.validation)
+        return self._validation
+
+    def set_test(self) -> tf.data.Dataset:
+        return self._wrap_tfdata(self.parent.set_test())
+
+    @property
+    def test(self) -> tf.data.Dataset:
+        if self._test is None:
+            self._test = self._wrap_tfdata(self.parent.test)
+        return self._test
+
+
 class _BatchDataBlob(_WrapDataBlob):
     """
     Batch the training, validation, and test sets.
@@ -697,12 +965,14 @@ class _BatchDataBlob(_WrapDataBlob):
     the batch size by the number of available replicas.
     """
 
-    def __init__(self, *, wraps, batch_size: int, with_tf_distribute: bool = False):
+    def __init__(
+        self, *, wraps: Any, batch_size: int, with_tf_distribute: bool = False
+    ):
         super().__init__(wraps=wraps)
         self._input_batch_size = batch_size
         self._with_tf_distribute = with_tf_distribute
         if self._with_tf_distribute:
-            self._final_batch_size = (
+            self._final_batch_size: int = (
                 self._input_batch_size
                 * tf.distribute.get_strategy().num_replicas_in_sync
             )
@@ -733,43 +1003,41 @@ class _LoadDataBlob(DataBlob):
     def __init__(
         self,
         *,
-        this_dataset_directory: str,
+        path: str,
         name: str,
         group_name: str,
         hyperparams: HyperparamsType,
     ):
         self.Hyperparams = hyperparams.__class__
         super().__init__(hyperparams=hyperparams)
-        self._this_dataset_directory = this_dataset_directory
+        self._path = path
         self._name = name
         self._group_name = group_name
 
     @classmethod
-    def load_from_directory(cls, this_dataset_directory):
-        metadata_path = os.path.join(this_dataset_directory, _METADATA_PICKLE_FILENAME)
+    def from_exact_path(cls, path: str) -> Union[DataBlob, DataFrameDataBlob]:
+        metadata_path = os.path.join(path, _METADATA_PICKLE_FILENAME)
         try:
             with open(metadata_path, "rb") as fh:
                 metadata = scalarstop.pickle.load(fh)
         except FileNotFoundError as exc:
-            raise DataBlobNotFound(this_dataset_directory) from exc
+            raise DataBlobNotFound(path) from exc
         name = metadata["name"]
         group_name = metadata["group_name"]
         hyperparams = metadata["hyperparams"]
         return cls(
-            this_dataset_directory=this_dataset_directory,
+            path=path,
             name=name,
             group_name=group_name,
             hyperparams=hyperparams,
         )
 
-    def _load_tfdata(self, subtype) -> tf.data.Dataset:
+    def _load_tfdata(self, subtype: str) -> tf.data.Dataset:
         """Load one of the :py:class:`tf.data.Dataset` s that we have saved."""
         try:
-            return _load_tfdata_dataset(
-                os.path.join(self._this_dataset_directory, subtype)
-            )
+            return _load_tfdata_dataset(os.path.join(self._path, subtype))
         except ElementSpecNotFound as exc:
-            raise TensorFlowDatasetNotFound(self._this_dataset_directory) from exc
+            raise TensorFlowDatasetNotFound(self._path) from exc
 
     def set_training(self) -> tf.data.Dataset:
         """An instance of the training set tf.data."""
@@ -788,9 +1056,7 @@ class _LoadDataFrameDataBlob(_LoadDataBlob, DataFrameDataBlob):
     """Loads a saved :py:class:`DataFrameDataBlob` from the filesystem."""
 
     def _set_subtype_dataframe(self, subtype: str) -> pd.DataFrame:
-        return pd.read_pickle(
-            os.path.join(self._this_dataset_directory, subtype, _DATAFRAME_FILENAME)
-        )
+        return pd.read_pickle(os.path.join(self._path, subtype, _DATAFRAME_FILENAME))
 
     def set_dataframe(self) -> pd.DataFrame:
         return pd.concat(
