@@ -30,7 +30,7 @@ import datetime
 import logging
 import sqlite3
 import urllib.parse
-from typing import Any, Dict, Optional, Sequence, Union
+from typing import Any, Dict, List, Literal, Optional, Sequence, Union
 
 import pandas as pd
 import sqlalchemy.dialects.postgresql
@@ -53,6 +53,7 @@ from sqlalchemy import (
 )
 from sqlalchemy import insert as default_insert
 from sqlalchemy import select, text
+from sqlalchemy.engine import Connection, Engine
 from sqlalchemy.exc import IntegrityError
 
 from scalarstop._datetime import utcnow
@@ -98,13 +99,13 @@ class _ModelMetadata:
     sort_metric_value: float
 
 
-def _enforce_list(value: Union[str, Sequence]):
+def _enforce_list(value: Union[str, Sequence[Any]]) -> List[Any]:
     if isinstance(value, str):
         return [value]
     return list(value)
 
 
-def _censor_sqlalchemy_url_password(url: str):
+def _censor_sqlalchemy_url_password(url: str) -> str:
     """
     Returns a SQLAlchemy URL without the password.
 
@@ -124,7 +125,7 @@ def _censor_sqlalchemy_url_password(url: str):
     return safe_url
 
 
-def _sqlite_json_enabled():
+def _sqlite_json_enabled() -> bool:
     """Return True if this Python installation supports SQLite3 JSON1."""
     connection = sqlite3.connect(":memory:")
     cursor = connection.cursor()
@@ -137,6 +138,22 @@ def _sqlite_json_enabled():
     return True
 
 
+def _flatten_hyperparam_results(results) -> List[Dict[str, str]]:
+    results_dicts = []
+    for row in results:
+        row_dict = dict(row)
+        dbh = row_dict.pop("datablob_hyperparams")
+        mth = row_dict.pop("model_template_hyperparams")
+        results_dicts.append(
+            dict(
+                **row_dict,
+                **{f"dbh__{key}": val for key, val in dbh.items()},
+                **{f"mth__{key}": val for key, val in mth.items()},
+            )
+        )
+    return results_dicts
+
+
 class _TrainStoreTables:
     """Manages our :py:class:`sqlalchemy.schema.Table` objects."""
 
@@ -263,7 +280,7 @@ class _TrainStoreTables:
         return self._metadata
 
     @property
-    def datablob(self):
+    def datablob(self) -> Table:
         """
         The :py:class:`sqlalchemy.schema.Table` used to store
         :py:class:`~scalarstop.DataBlob` objects.
@@ -271,7 +288,7 @@ class _TrainStoreTables:
         return self._datablob_table
 
     @property
-    def model_template(self):
+    def model_template(self) -> Table:
         """
         The :py:class:`sqlalchemy.schema.Table`used to store
         :py:class:`~scalarstop.ModelTemplate` objects.
@@ -279,7 +296,7 @@ class _TrainStoreTables:
         return self._model_template_table
 
     @property
-    def model(self):
+    def model(self) -> Table:
         """
         The :py:class:`sqlalchemy.schema.Table` used to store
         :py:class:`~scalarstop.model.Model` objects.
@@ -287,7 +304,7 @@ class _TrainStoreTables:
         return self._model_table
 
     @property
-    def model_epoch(self):
+    def model_epoch(self) -> Table:
         """
         The :py:class:`sqlalchemy.schema.Table` used to store
         metrics from model epochs.
@@ -305,7 +322,7 @@ class TrainStore:
         filename: str,
         table_name_prefix: Optional[str] = None,
         echo: bool = False,
-    ):
+    ) -> "TrainStore":
         """
         Use a SQLite3 database file on the local filesystem as the train store.
 
@@ -391,7 +408,7 @@ class TrainStore:
         return f"<sp.TrainStore {self._connection_string_no_password}>"
 
     @property
-    def table(self):
+    def table(self) -> _TrainStoreTables:
         """
         References to the :py:class:`sqlalchemy.schema.Table` objects
         representing our database tables.
@@ -407,7 +424,7 @@ class TrainStore:
         return self._table
 
     @property
-    def engine(self):
+    def engine(self) -> Engine:
         """
         The currently active :py:class:`sqlalchemy.engine.Engine`.
 
@@ -417,7 +434,7 @@ class TrainStore:
         return self._engine
 
     @property
-    def connection(self):
+    def connection(self) -> Connection:
         """
         The currently active :py:class:`sqlalchemy.engine.Connection`.
 
@@ -426,7 +443,9 @@ class TrainStore:
         """
         return self._connection
 
-    def _insert(self, *, table, values, index_elements=None, ignore_existing: bool):
+    def _insert(
+        self, *, table, values, index_elements=None, ignore_existing: bool
+    ) -> None:
         if ignore_existing:
             if self._engine.name == "sqlite":
                 with self.connection.begin():
@@ -452,11 +471,11 @@ class TrainStore:
             with self.connection.begin():
                 self.connection.execute(default_insert(table).values(**values))
 
-    def _as_pandas(self, stmt):
+    def _as_pandas(self, stmt) -> pd.DataFrame:
         with self.connection.begin():
             return pd.read_sql_query(sql=stmt, con=self.connection)
 
-    def insert_datablob(self, datablob, *, ignore_existing: bool = False):
+    def insert_datablob(self, datablob, *, ignore_existing: bool = False) -> None:
         """
         Logs the :py:class:`~scalarstop.datablob.DataBlob` name, group name,
         and hyperparams to the :py:class:`TrainStore`.
@@ -483,7 +502,12 @@ class TrainStore:
         )
 
     def insert_datablob_by_str(
-        self, *, name: str, group_name: str, hyperparams, ignore_existing: bool = False
+        self,
+        *,
+        name: str,
+        group_name: str,
+        hyperparams: Any,
+        ignore_existing: bool = False,
     ):
         """
         Logs the :py:class:`~scalarstop.datablob.DataBlob` name, group
@@ -750,7 +774,7 @@ class TrainStore:
         datablob_name: str,
         model_template_name: str,
         ignore_existing: bool = False,
-    ):
+    ) -> None:
         """
         Logs the :py:class:`~scalarstop.model.Model` name,
         :py:class:`~scalarstop.datablob.DataBlob`, and
@@ -865,7 +889,7 @@ class TrainStore:
         stmt = stmt.order_by(self.table.model.c.model_last_modified)
         return stmt
 
-    def _query_model_by_epoch_stmt(  # pylint: disable=too-many-branches
+    def _query_model_by_epoch_stmt(
         self,
         *,
         metric_name: Optional[str] = None,
@@ -875,6 +899,7 @@ class TrainStore:
         model_template_name: Optional[Union[str, Sequence[str]]] = None,
         model_template_group_name: Optional[Union[str, Sequence[str]]] = None,
         limit: Optional[int] = None,
+        return_other_metrics: bool = True,
     ):
         where_conditions = []
 
@@ -910,7 +935,6 @@ class TrainStore:
             self.table.model.c.model_name,
             self.table.model.c.model_class_name,
             self.table.model.c.model_last_modified,
-            self.table.model_epoch.c.model_epoch_metrics,
             self.table.datablob.c.datablob_name,
             self.table.datablob.c.datablob_group_name,
             self.table.datablob.c.datablob_hyperparams,
@@ -918,6 +942,10 @@ class TrainStore:
             self.table.model_template.c.model_template_group_name,
             self.table.model_template.c.model_template_hyperparams,
         ]
+
+        if return_other_metrics:
+            columns.append(self.table.model_epoch.c.model_epoch_metrics)
+
         group_by_columns = columns.copy()
         sort_metric_value = None
         if metric_name and metric_direction:
@@ -980,7 +1008,7 @@ class TrainStore:
         datablob_group_name: Optional[Union[str, Sequence[str]]] = None,
         model_template_name: Optional[Union[str, Sequence[str]]] = None,
         model_template_group_name: Optional[Union[str, Sequence[str]]] = None,
-    ):
+    ) -> pd.DataFrame:
         """
         Returns a :py:class:`pandas.DataFrame` listing ALL of the rows in the
         :py:class:`~scalarstop.model.Model` table.
@@ -1013,7 +1041,6 @@ class TrainStore:
                 :py:class:`~scalarstop.model_template.ModelTemplate`
                 group name or a list of group names to select.
         """
-        results_dicts = []
         stmt = self._query_model_stmt(
             datablob_name=datablob_name,
             datablob_group_name=datablob_group_name,
@@ -1022,22 +1049,96 @@ class TrainStore:
         )
         with self.connection.begin():
             results = self.connection.execute(stmt)
-            for row in results:
-                row_dict = dict(row)
-                dbh = row_dict.pop("datablob_hyperparams")
-                mth = row_dict.pop("model_template_hyperparams")
-                results_dicts.append(
-                    dict(
-                        **row_dict,
-                        **{f"dbh__{key}": val for key, val in dbh.items()},
-                        **{f"mth__{key}": val for key, val in mth.items()},
-                    )
-                )
+            results_dicts = _flatten_hyperparam_results(results)
+        return pd.DataFrame(results_dicts)
+
+    def list_models_grouped_by_epoch_metric(
+        self,
+        *,
+        metric_name: str,
+        metric_direction: str,
+        datablob_name: Optional[Union[str, Sequence[str]]] = None,
+        datablob_group_name: Optional[Union[str, Sequence[str]]] = None,
+        model_template_name: Optional[Union[str, Sequence[str]]] = None,
+        model_template_group_name: Optional[Union[str, Sequence[str]]] = None,
+    ) -> pd.DataFrame:
+        """
+        Returns a :py:class:`pandas.DataFrame` listing ALL of the rows in the
+        :py:class:`~scalarstop.model.Model` table AND a metric from
+        the model's best-performing epoch.
+
+        You provide this method with a model epoch metric name
+        and whether to maximize or minimize this, and then
+        it returns all of the models and the best metric value.
+
+        Note that you can provide either ``datablob_name`` or
+        ``datablob_group_name``, but not both.
+
+        Similarly, you can provide either ``model_template_name``
+        or ``model_template_group_name``, but not both.
+
+        Args:
+            metric_name: The name of one of the metrics
+                tracked when training a model. This might be a value
+                like ``"loss"`` or ``"val_accuracy"``.
+
+            metric_direction: Set this to ``"min"`` if the metric
+                you picked in ``metric_name`` is a value where
+                lower values are better--such as ``"loss"``.
+                Set this to ``"max"`` if higher values of your
+                metric are better--such as ``"accuracy"``.
+
+            datablob_name: Either a single
+                :py:class:`~scalarstop.datablob.DataBlob`
+                name or a list of names to select.
+
+            datablob_group_name: Either a single
+                :py:class:`~scalarstop.datablob.DataBlob`
+                group name or a list of group names to select.
+
+            model_template_name: Either a single
+                :py:class:`~scalarstop.model_template.ModelTemplate`
+                name or a list of names to select.
+
+            model_template_group_name: Either a single
+                :py:class:`~scalarstop.model_template.ModelTemplate`
+                group name or a list of group names to select.
+
+        Returns a :py:class:`pandas.DataFrame` with the following
+        columns:
+
+        * ``model_name``
+        * ``model_class_name``
+        * ``model_last_modified``
+        * ``datablob_name``
+        * ``datablob_group_name``
+        * ``model_template_name``
+        * ``model_template_group_name``
+        * ``sort_metric_value``
+        * :py:class:`~scalarstop.model_template.ModelTemplate`
+          hyperparameter names prefixed with ``mth__``
+        * :py:class:`~scalarstop.datablob.DataBlob`
+          hyperparameter names prefixed with ``dbh__``
+
+        """
+        stmt = self._query_model_by_epoch_stmt(
+            metric_name=metric_name,
+            metric_direction=metric_direction,
+            datablob_name=datablob_name,
+            datablob_group_name=datablob_group_name,
+            model_template_name=model_template_name,
+            model_template_group_name=model_template_group_name,
+            return_other_metrics=False,
+        )
+        with self.connection.begin():
+            results_dicts = _flatten_hyperparam_results(
+                self.connection.execute(stmt).fetchall()
+            )
         return pd.DataFrame(results_dicts)
 
     def insert_model_epoch(
         self, *, epoch_num: int, model_name: str, metrics, ignore_existing: bool = False
-    ):
+    ) -> None:
         """
         Logs a new epoch for a :py:class:`~scalarstop.model.Model`
         to the :py:class:`TrainStore`.
@@ -1070,7 +1171,7 @@ class TrainStore:
             ignore_existing=ignore_existing,
         )
 
-    def bulk_insert_model_epochs(self, model):
+    def bulk_insert_model_epochs(self, model) -> None:
         """
         Insert a list of :py:class:`~scalarstop.model.Model` epochs at once.
 
@@ -1142,7 +1243,7 @@ class TrainStore:
     def list_model_epochs(
         self,
         model_name: Optional[Union[str, Sequence[str]]] = None,
-    ):
+    ) -> pd.DataFrame:
         """
         Returns a :py:class:`pandas.DataFrame` listing
         :py:class:`~scalarstop.model.Model` epochs.
@@ -1171,7 +1272,7 @@ class TrainStore:
                 )
         return pd.DataFrame(results_dicts)
 
-    def get_current_epoch(self, model_name: str):
+    def get_current_epoch(self, model_name: str) -> int:
         """
         Returns how many epochs a given :py:class:`~scalarstop.model.Model` has been
         trained for.
@@ -1288,10 +1389,10 @@ class TrainStore:
         )
         return the_dataclass
 
-    def __enter__(self):
+    def __enter__(self) -> "TrainStore":
         return self
 
-    def close(self):
+    def close(self) -> None:
         """
         Close the database connection.
 
@@ -1299,6 +1400,6 @@ class TrainStore:
         """
         self.connection.close()
 
-    def __exit__(self, exc_type, exc_value, exc_traceback):
+    def __exit__(self, exc_type, exc_value, exc_traceback) -> Literal[False]:
         self.close()
         return False
